#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['docopt_sh']

package_data = \
{'': ['*']}

install_requires = \
['docopt', 'termcolor']

extras_require = \
{'test': ['pytest', 'pytest-xdist', 'pytest-pep8']}

entry_points = \
{'console_scripts': ['docopt.sh = docopt_sh.__main__:main']}

setup(name='docopt-sh',
      version='0.9.10',
      description='docopt.sh - Bash argument parser generator.',
      author='Anders Ingemann',
      author_email='anders@ingemann.de',
      url='https://github.com/andsens/docopt.sh',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.5',
     )
